/* 
 * wsdlpull - A C++ parser  for WSDL  (Web services description language)
 * Copyright (C) 2005-2007 Vivek Krishna
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */
#include <sstream>
#include <schemaparser/SchemaValidator.h>
using namespace std;

/*
  This class validates an incoming Xml instance against a given type
  whose schema has been processed by a given SchemaParser instance

*/
SchemaValidator::SchemaValidator(const SchemaParser * sp)
{
  sParser = sp;
}

SchemaValidator::~SchemaValidator()
{
}


/*
  Main entry method for validation
  Inputs
  1. XmlStream ,xpp.getName() muct return the name of the  type
  which must be validated
  2. typeId of the type against which this stream must be
  validated against
  3.An Input type container  (default  0)
*/
TypeContainer *
SchemaValidator::validate(XmlPullParser * xpp, 
			  int typeId,
			  TypeContainer * ipTc)
{
  try{  
    TypeContainer *t;
    string elemName = xpp->getName();

#ifdef LOGGING
    std::cout<<elemName<<std::endl;
#endif

    if (!ipTc)
      t = new TypeContainer(typeId, sParser);
    else
      t = ipTc;

    if (t->getTypeId() != typeId)
      error("Fatal error ,container's type is not same as the validated type",xpp);

    // a schema definition inside an instance
    if (typeId == Schema::SCHEMA){
      
      SchemaParser * ssParser = new SchemaParser(xpp);
      if (!ssParser->parseSchemaTag()){
	
	return 0;
      }
      return t;
    }

    //ignore ANY
    if (typeId == Schema::ANY){
      
      xpp->skipSubTree();
      return t;
    }
    
    if (sParser->getType(typeId) == 0
	|| sParser->getType(typeId)->isSimple()) {

      //simple type validation
      string val;
      xpp->nextToken();
      if (xpp->getEventType() == XmlPullParser::TEXT ||
	  xpp->getEventType() == XmlPullParser::ENTITY_REF){

	val = xpp->getText();

	xpp->nextToken();
	while (xpp->getEventType() == XmlPullParser::ENTITY_REF ||
	       xpp->getEventType() == XmlPullParser::TEXT){

	  val += xpp->getText();
	  xpp->nextToken();
	    
	}
	validate(val, typeId, t);
      }
      else{
	//text was probably empty,nevertheless create  a type container 
	validate(val, typeId, t);
      }
      if (xpp->getEventType() == XmlPullParser::END_TAG)
	{
	  if (xpp->getName() != elemName)
	    error("Syntax error "+elemName,xpp);
	}
      else
	error("Expected a closing tag for " + elemName,xpp);
    }
    else {

      /*
	Perform Validation of Complex types
	Check for
	1.Is the tag name correct (this has to be right !!)
	2.Attributes ,if any should be valid
	3.Are there any mandatory (#required) attributes
	4. Validate its content model 
	6.Return the type container which has the
	correctly filled in values

      */
      const ComplexType *ct =
	static_cast<const ComplexType *>(sParser->getType(typeId));
	
      const ComplexType * bt = 0;
      TypeContainer * btCnt = 0;
      if (ct->getBaseTypeId()!=Schema::ANYTYPE) {

	bt = static_cast<const ComplexType*>
	  (sParser->getType(ct->getBaseTypeId()));
	btCnt = t->getBaseTypeContainer(true);
      }
	
      int attcnt = xpp->getAttributeCount();
	
      for (int i = 0; i < attcnt; i++) {
	  
	std::string attName = xpp->getAttributeName(i);
	std::string attVal = xpp->getAttributeValue("", attName);
	std::string attNsp = xpp->getAttributeNamespace(i);
	if (!attNsp.empty() && attNsp != sParser->getNamespace())
	  continue;

	const Attribute*at = 0;
	TypeContainer *atCnt = 0;
	at = ct->getAttribute(attName);

	if (!at && bt){
	  at  = bt->getAttribute(attName);
	  if (at)
	    atCnt = btCnt->getAttributeContainer(attName, true);
	}
	else{
	  atCnt = t->getAttributeContainer(attName, true);
	}

	if (!at)
	  error("Unkown attribute \"" + attName +  "\"",xpp);

	validate(attVal, at->getType(), atCnt);
      }
	
      //see if some required attributes are missing
      checkAttributeOccurence(ct,xpp);
      if (bt)
	checkAttributeOccurence(bt,xpp);


      if (ct->getContentModel() == Schema::Simple)
	{
	  //complex types with a simple content model

	  string val;
	  xpp->nextToken();
	  if (xpp->getEventType() == xpp->TEXT){
	    val = xpp->getText();
	    validate(val, ct->getContentType(), t);
	    xpp->nextTag();
	  }
	  else{
	    //text was probably empty,nevertheless create  a type container 
	    validate(val, ct->getContentType(), t);
	  }

	  if (xpp->getEventType() == XmlPullParser::END_TAG)
	    {
	      if (xpp->getName() != elemName)
		error("Syntax error",xpp);
	    }
	  else
	    error("Expected a closing tag for " + elemName,xpp);
	}
      else if (ct->getContentModel() == Schema::Complex){
	//a complex type with complex content model
	ContentModel* cm=ct->getContents();
	if(cm)
	  validateContentModel(xpp,
			       cm,
			       t->getChildContainer(cm,true),
			       elemName);
      }
      else{
	// a complex type with mixed content model.no support yet
      }
    }
    return t;
  }catch (SchemaParserException spe){
    if(xpp){
      spe.line=xpp->getLineNumber();
      spe.col=xpp->getColumnNumber();    
      throw spe;
    }
  }
  return 0;
}

TypeContainer*
SchemaValidator::validateContentModel(XmlPullParser * xpp, 
                                      ContentModel* cm,
                                      TypeContainer * ipTc,
                                      const string & elemName,
				      bool nested)
{
  ContentModel::ContentsIterator cit_b=cm->begin();
  ContentModel::ContentsIterator cit_e=cm->end();
  ContentModel::ContentsIterator ci=cit_e;
  
  for (ci=cit_b;ci!=cit_e;ci++){
    if(ci->second==ContentModel::Particle)
      ci->first.e->nOccurrences=0;
  }
  ci=cit_b;
  switch (cm->getCompositor())
    {
    case Schema::All:
      {
	do
	  {
	    if (!nested)
	      xpp->nextTag();
	    if (xpp->getEventType() == XmlPullParser::END_TAG)
	      {
		if (xpp->getName() == elemName)
		  break;
		while (xpp->getEventType() != XmlPullParser::START_TAG)
		  xpp->nextTag();
	      }
	    //All cannot have another content model inside like group/choice etc
 
	    if(!findElement(cit_b,cit_e,xpp->getName(),ci))
	      error("Could not find element " +xpp->getName()+" in "+elemName,xpp);
	    ci->first.e->nOccurrences++;

	    validate(xpp, ci->first.e->getType(),
		     ipTc->getChildContainer(ci->first.e->getName(), true));
	    //ipTc->getChildContainer(xpp->getName(), true));
	  }
	while (true);

	/*
	  check for occurrence constraints
	*/
	for (ci=cit_b;ci!=cit_e;ci++){
	  if(ci->second==ContentModel::Particle && 
	     (ci->first.e->nOccurrences<ci->first.e->min()||
	      ci->first.e->nOccurrences>ci->first.e->max()))
	    error(ci->first.e->getName()+" did not meet occurrence constraints",xpp);
	}
	
	break;
      }
    case Schema::Sequence:
      {
	do
	  {
	    if (!nested)
	      xpp->nextTag();
	    
	    if(xpp->getEventType() != XmlPullParser::END_TAG){

	      if(ci->second==ContentModel::Particle &&
		 xpp->getName()!=ci->first.e->getName() &&
		 ci->first.e->getName()!="*") 
		ci++;
	      //check for multiple occurrences of the same element
	      //if the same element occurrs more than once dont advance iterator ci

	      //TODO multiple occurrences of a content model like sequence
	      //cant be validated as yet
	    }else {
	      
	      if (xpp->getName() == elemName)
		break;
	      if(ci==cit_e)
		break;
	      while (xpp->getEventType() != XmlPullParser::START_TAG)
		xpp->nextTag();
	    }

	    if(ci->second==ContentModel::Container){
            
	      validateContentModel(xpp,ci->first.c,
				   ipTc->getChildContainer(ci->first.c,true),
				   elemName,true);
	      ci++;
	    }else{

	      if(!findElement(ci,cit_e,xpp->getName(), ci) &&
		 !cm->anyContents())
		error("Could not find element " +xpp->getName()+" in "+elemName,xpp);
	      
	      ci->first.e->nOccurrences++;
	      validate(xpp,ci->first.e->getType(),
		       ipTc->getChildContainer(ci->first.e->getName(), true));
	      //ipTc->getChildContainer(xpp->getName(), true));
	    }
	  }
	while (true);

	/*
	  check for occurrence constraints
	*/
	for (ci=cit_b;ci!=cit_e;ci++){
	  if(ci->second==ContentModel::Particle && 
	     (ci->first.e->nOccurrences<ci->first.e->min()||
	      ci->first.e->nOccurrences>ci->first.e->max()))
	    error(ci->first.e->getName()+" did not meet occurrence constraints",xpp);
	}
	break;
      }      
    case Schema::Choice:
      {

	if (!nested)
	  xpp->nextTag();
	if(ci->second==ContentModel::Container){
        
	  validateContentModel(xpp,ci->first.c,
			       ipTc->getChildContainer(ci->first.c,true),
			       elemName,true);
	  ci++;
	  
        
	}else{
        
	  if(!findElement(ci, cit_e, xpp->getName(), ci))
	    error("Could not find element " +xpp->getName()+" in "+elemName,xpp);
	  ci->first.e->nOccurrences++;
	  validate(xpp, ci->first.e->getType(),
		   ipTc->getChildContainer(ci->first.e->getName(), true));
	  //   ipTc->getChildContainer(xpp->getName(), true));
	}
         
	/*
	 * Only one of the choices is allowed
	 */
	
	/*
	 * check for occurrence constraints
	 */
	if(ci->second==ContentModel::Particle && 
	   (ci->first.e->nOccurrences<ci->first.e->min()||
	    ci->first.e->nOccurrences>ci->first.e->max()))
	  error(ci->first.e->getName()+"did not meet occurrence constraints",xpp);
      
	break;
      }
    }
  /*
   * reset occurence ocunters
   */
  for (ci=cit_b;ci!=cit_e;ci++){
    
    if(ci->second==ContentModel::Particle)
      ci->first.e->nOccurrences=0;
  }
  return ipTc;
}
  


/*
 *  This method validates all supported simple types
 *  Both native atomic types and schema defined 
 */

TypeContainer * 
SchemaValidator::validate(void* value ,
                          int typeId,
                          TypeContainer * ipTc)
{
  
  int basetype = sParser->getBasicContentType(typeId);
  
  const XSDType * pType = sParser->getType(typeId);
  if (pType && !pType->isSimple()){
    
    return 0;
  }
  const SimpleType  *st = static_cast<const SimpleType*>(pType);
  
  //check for the validity of the value
  //if available also check against restrictions in the schema

  if (!ipTc)
    ipTc = new TypeContainer(typeId, sParser);

  if (st && (st->isList() || st->isUnion())){
	    
    std::string val = *((std::string*)value);
    ipTc->setValue(val,validateListOrUnion(st,val));
    return ipTc;
  }
  switch (basetype)
    {
    case Schema::INTEGER:
    case Schema::INT:
      {
	int x= *((int*)value);
	if (!st) {
	  ipTc->setValue(x);
	}
	else{

	  ipTc->setValue(x,st->isValidInt(x));
	}
	break;
      }
    case Schema::BYTE:
      {
	char c= *((char*)value);
	ipTc->setValue(c);
      }
      break;
    case Schema::FLOAT:
      {
	float f = *((float*)value);
	if (!st) {
	  
	  ipTc->setValue(f);
	  
	}else{
	  
	  ipTc->setValue(f,st->isValidFloat(f));
	}
	break;
      }
    case Schema::DOUBLE:
    case Schema::DECIMAL:
      {
	double db = *((double*)value);
	ipTc->setValue(db);
      }
      break;
    case Schema::LONG:
      {
	long l = *((long*)value);
	ipTc->setValue(l);
      }
      break;
    case Schema::POSINT:
    case Schema::ULONG:
      {
	unsigned long ul= *((unsigned long*)value);
	ipTc->setValue(ul);
      }
      break;
    case Schema::BOOLEAN:
      {
	bool b  = *((bool*)value);
	ipTc->setValue(b);
	break;
      }
    case Schema::QNAME:
      {
	Qname q  = *((Qname* )value);
	ipTc->setValue(q);
      }
      break;
    case Schema::STRING:
    default: 
      {
	std::string val = *((std::string* )value);
	if (!st) {
	  
	  ipTc->setValue(val);
	}
	else{
	  ipTc->setValue(val,st->isValidString(val));
	}
      }
      break;
    }

  return  ipTc;
}

/*
 *  This method validates all supported simple types
 *  Both native atomic types and schema defined 
 */

TypeContainer *
SchemaValidator::validate(const string & val,
			  int typeId,
			  TypeContainer *ipTc)
{

  int basetype = sParser->getBasicContentType(typeId);
  if (basetype == Schema::INVALID) {
    
    return 0;
  }
  
  const XSDType * pType = sParser->getType(typeId);
  if (pType && !pType->isSimple()){
    
    return 0;
  }
  const SimpleType  *st = static_cast<const SimpleType*>(pType);
  
  //check for the validity of the value
  //if available also check against restrictions in the schema

  if (!ipTc)
    ipTc = new TypeContainer(typeId, sParser);
  ipTc->setValAsString(val);

  while(ipTc->isValueValid()){
    
    extractSimpleType(val, basetype, ipTc, st);

    
    if(!st || (st && (st->isList() || st->isUnion()))){
      
      break;
      //if we validated an atomic type we are done
      //if we just validated a list or union,there is no need 
      //to continue checking base types
    }

    if (!sParser->isBasicType(st->getBaseTypeId())){
      
      st=static_cast<const SimpleType*>(sParser->getType(st->getBaseTypeId()));
    }
    else{
      st = 0;
    }
  }
  return ipTc;
}


void 
SchemaValidator::extractSimpleType(const std::string & val,
				   int basetype,
				   TypeContainer * ipTc,
				   const SimpleType * st)
{

  if (st && (st->isList() || st->isUnion())){
	    
    ipTc->setValue(val,validateListOrUnion(st,val));
    return;
  }
  
  istringstream istr(val);
  int x;
  double db;
  long l;
  char c;
  unsigned long ul;
  float f;
  Qname *q = 0;
#ifdef LOGGING
  std::cout<<val<<std::endl;
#endif
  switch (basetype)
    {
    case Schema::INTEGER:
    case Schema::INT:
      {
	istr >> x;
	if (!st) {
	  ipTc->setValue(x,!istr.fail());
	}
	else{

	  ipTc->setValue(x,!istr.fail() && st->isValidInt(x));
	}
	break;
      }
    case Schema::BYTE:
      istr >> c;
      ipTc->setValue(c,!istr.fail());
      break;
    case Schema::FLOAT:
      {
	istr >> f;
	if (!st) {
	  ipTc->setValue(f,!istr.fail());
	}else{
	  ipTc->setValue(f,!istr.fail() && st->isValidFloat(f));
	}
	break;
      }
    case Schema::DOUBLE:
    case Schema::DECIMAL:
      istr >> db;
      ipTc->setValue(db,!istr.fail());
      break;
    case Schema::LONG:
      istr >> l;
      ipTc->setValue(l,!istr.fail());
      break;
    case Schema::POSINT:
    case Schema::ULONG:
      istr >> ul;
      ipTc->setValue(ul,!istr.fail());
      break;
    case Schema::BOOLEAN:
      {

	if(val=="true" ||
	   val=="yes" ||
	   val=="1")
	
	  ipTc->setValue(true);
	else
	  ipTc->setValue(false);
	break;
      }
    case Schema::QNAME:
      q = new Qname(val);
      ipTc->setValue(*q);
      break;
    case Schema::STRING:
    default: 
      {
	if (!st) {
	  ipTc->setValue(val);
	}
	else{
	  if (basetype == Schema::STRING)
	    ipTc->setValue(val,st->isValidString(val));
	  else
	    ipTc->setValue(val);//other types such as date for which no validation is done
	}
      }
      break;
    }
}

/*
 * This function validates a string as a list or union 
 * for the simple type 
 */

bool
SchemaValidator::validateListOrUnion(const SimpleType* st,
				     const std::string &val)
{
  if (st->isList()){
   
    size_t s = 0;

    while(s < val.length()){
      while(val[s]==' ')s++;
      std::string t = val.substr(s,val.find(' ',s)-s);
      TypeContainer * tc = validate(t,st->getBaseTypeId());
      if (!(tc && tc->isValueValid()))
	return false;
      s+=t.length()+1;
    }
    return true ;
  
  }else if (st->isUnion()){
    
    std::list<int>::const_iterator it= st->unionTypes()->begin();
    while (it!=st->unionTypes()->end()){
      TypeContainer * tc = validate(val,*it);
      if (tc && tc->isValueValid())
	return true;
    }
    return false;
  }
  else{
    return false;
  }
}
/*
 * This function searches for a child element in a complex type 
 * The iterator pElem is set to point to the found element
 * rewind controls whether a search is to be done by resetting pElem 
 * to the begining of the list 
 * Returns true if element is found else false 
 */

bool
SchemaValidator::findElement(ContentModel::ContentsIterator start, 
			     ContentModel::ContentsIterator end, 
			     std::string name,
			     ContentModel::ContentsIterator & found)
{
  for (ContentModel::ContentsIterator ci=start;
       ci!=end;
       ci++){
    if(ci->second==ContentModel::Particle){
      if(ci->first.e->getName()==name ||
	 ci->first.e->getName() == "*")
        {
          found=ci;
          return true;
        }
    }
  }
  return false;
}

void SchemaValidator::error(const std::string& mesg,XmlPullParser* xpp)
{
    
  SchemaParserException spe(mesg + "\nError validating schema instance\n");
  if(xpp){
    spe.line=xpp->getLineNumber();
    spe.col=xpp->getColumnNumber();
  }
  throw spe;
}


bool
SchemaValidator::checkAttributeOccurence(const ComplexType* ct ,
					 XmlPullParser* xpp)
{

  if (ct->getNumAttributes() > 0)
    {
      for (int i = 0; i < ct->getNumAttributes(); i++)
	{
	  const Attribute*at = ct->getAttribute(i);
              
	  /*
	    Check for the correctness of each attribute
	  */
	  string attVal = xpp->getAttributeValue("", at->getName());
	  if (attVal.empty())
	    {
	      if (at->isRequired())
		error("Required attribute \"" + at->getName() +
		      "\" missing or empty",xpp);

	      else
		continue;
	    }
	}
    }
  return true;
}
//TODO validation of <any>
//TODO validation of base64binary,hexBinary
//TODO validation of types derived by extension/restriction of complex types
